package ASCOPE::Article::Collection;
use strict;

# $Id: Collection.pm,v 1.5 2004/07/10 17:04:12 asc Exp $

=head1 NAME

ASCOPE::Article::Collection - Aaron's widget for munging a specialized subset of the DocBook format

=head1 SYNOPSIS

 use ASCOPE::Article::Collection;
 my $collection = ASCOPE::Article::Collection->new("/path/to/articles.xml");

 $collection->mk_html({htroot => "/htdocs/news",
                       uri    => "http://127.0.0.1/news",
                       flush  => 1});

=head1 DESCRIPTION

Aaron's widget for munging a specialized subset of the DocBook format used to
record one or more articles.

=head1 DOCUMENT STRUCTURE

Book widgets contain the following elements :

 + book
   + bookinfo
     - title
     - pubdate
   + part (+)
     + partinfo
       - title
     + xref (+)
       @linkend
     + article (+)

Article widgets contain the following elements :

 + article
   @id
   + articleinfo
     - title
     + keywordset (or not)
       - keyword
     + indexterm (or not)
       @ type
       - primary
     + authorgroup
       - corpauthor
       OR
       + author
        - surname
        - firstname
        - othername (or not)
     + abstract
       - para
     - pubdate
     + copyright
       - year
       - holder
   - para (+)

=head1 TOPICS, KEYWORDS, INDEXTERMS AND RELATIONS

Individual articles may contain one or more I<keywordset> or I<indexterm>
thingies. When these relations are indexed in any of the methods below,
they will be returned with the following labels :

=head2 keywords

A keyword is labeled as an "idea".

=head2 indexterms

Index terms are labeled as follows - where the type attribute is :

=over 4

=item * B<org>

"organization"

=item * B<geo>

"place"

=item * B<per>

"person"

=back

=cut

$ASCOPE::Article::Collection::VERSION = '1.1';

use XML::LibXML;
use File::Path;
use IO::File;
use CGI qw(:html);
use DirHandle;
use Memoize;
use Digest::MD5 qw (md5_hex);

&memoize("_normalize");

# Need to finish moving all the XPath up here...
use constant XPATH_ARTICLES => "child::node()[name() = 'article' or name() = 'xref']";

use constant RELS_INDEXTERMS => ("per","org","geo");
use constant RELS_KEYWORDS   => ("keywords");

use constant RELS_LABELS => {geo      => "place",
			     org      => "organization",
			     per      => "person",
			     keywords => "idea"};

use constant RELS_SHAPES => {geo      => "box",
			     keywords => "box",
			     org      => "box",
			     per      => "box"};

use constant RELS_COLOURS => {geo      => "pink",
			      keywords => "yellow",
			      org      => "orange",
			      per      => "cyan"};

=head1 PACKAGE METHODS

=cut

=head2 __PACKAGE__->new($path)

Returns a I<ASCOPE::Article::Collection> object. Woot!

=cut

sub new {
  my $pkg  = shift;
  my $path = shift;

  my $parser = XML::LibXML->new();
  my $doc    = $parser->parse_file($path);

  return bless {'__doc'=>$doc,'__path'=>$path},$pkg;
}

=head1 OBJECT METHODS

=cut

sub path {
    return $_[0]->{'__path'};
}

=head2 $obj->pubdate()

Returns a string.

=cut

sub pubdate {
  my $self = shift;
  $self->{'__date'} ||= $self->{'__doc'}->find(qq(/book/bookinfo/pubdate));
  return $self->{'__date'};
}

=head2 $pkg->mk_html(\%args)

=over 4

Render a collection of articles as an HTML tree.

=item * B<htroot>

The root path to a directory where files are created.

=item * B<uri>

String. 

The root URI for links.

=item * B<flush>

Boolean.

Remove old files and directories.

=item * B<verbose>

Boolean.

=back

Returns true or false.

=cut

sub mk_html {
  my $self = shift;
  my $args = shift;

  my $htroot = $args->{htroot};
  my $uri    = $args->{uri};

  if ((-d $htroot) && ($args->{flush})) {
    my $dh = DirHandle->new($htroot);

    foreach my $subdir ($dh->read()) {
      next if ($subdir =~ /^\./);
      rmtree("$htroot/$subdir", $args->{verbose});
    }
  }

  else { mkpath($htroot,$args->{verbose}); }

  # Create an index page for the day 

  my $title = $self->{'__doc'}->find(qq(/book/bookinfo/title));

  my $index = IO::File->new(">$htroot/index.html");
  $index->print(&start_html(-title=>$title));
  $index->print(&h1($title));

  #

  foreach my $section ($self->{'__doc'}->findnodes("/book/part")) {

    # The widget that generates the DocBook is SAX
    # driven and generates a 'part' (section) event
    # even if there are no articles for that date.

    my @articles = $section->findnodes(XPATH_ARTICLES);
    next unless (@articles);

    # Munge the section title

    my $sectname = $section->find(qq(partinfo/title));
    &_sectname(\$sectname);
    
    # Create directory root in which
    # we'll write articles.

    mkpath("$htroot/$sectname");

    # More happy HTML

    $index->print(&h2(&a({href=>"$uri/$sectname"},$sectname)));

    # Create an index page for this section

    my $sect = IO::File->new(">$htroot/$sectname/index.html");
    $sect->print(&start_html(-title=>"$title $sectname"));
    $sect->print(&h1($sectname));

    foreach my $article (@articles) {

      # If we're dealing with an xref then we will
      # need to make sure that the parent ($sectname)
      # for the article points to the source and not
      # the current section we're iterating through.

      # Rather than do some fancy-pants local() kung-fu
      # we'll just set up a 'parent' variable for this
      # block and get on with it.
      my $parent = $sectname;

      if ($article->nodeName() eq "xref") {
	my $ref = $article->getAttribute(qq(linkend));

	$article = ($self->{'__doc'}->findnodes(qq(*//article[\@id='$ref'])))[0];

	if (! $article) {
	  warn "Unable to find article: $ref\n";
	  next;
	}

	# Update the 'parent' so that filenames and
	# uris are generated correctly and point to
	# the source article

	$parent = $article->find(qq(parent::node()/partinfo/title));
	&_sectname(\$parent);
      }

      # Get the ID for the article (this is actually
      # an MD5 hex of the uri for the article.

      my $id  = $article->getAttribute(qq(id));
      my $md5 = md5_hex($id);

      # Pull out the articleinfo and create some
      # variables that we'll use below.

      my $info = &_articleinfo($article);

      if (! $info) {
	$sect->print(&p("Unable to find any articleinfo for '$id'\n"));
	next;
      }

      #

      my $title  = $info->find(qq(title));
      my $author = &_author($info);

      my $uri   = "$uri/$parent/$md5.html";

      # List the article on the main page.

      $index->print(&a({href=>$uri},$title));
      $index->print("&nbsp;$author");
      $index->print(&br());

      # List the article and its abstract
      # on the section page.

      $sect->print(&a({href=>$uri},$title));
      $sect->print(&br());
      $sect->print(&p($info->find(qq(abstract/para))));

      # Create the actual article HTML page.
      # Assuming that we're not dealing with
      # an reference, in which case it's already
      # been written to disk.

      if (! -f "$htroot/$parent/$md5.html") {
	my $text = IO::File->new(">$htroot/$parent/$md5.html");
	$text->print(&start_html(-title=>$title),
		     &h1($title),
		     &p($author));

	map {
	    $text->print(&p($_));
	} @{ &_cleanarticle($article) };

	$text->print(&a({href=>$id},$id));
	$text->print(&end_html());
	$text->close();
      }

    }

    # Finish up the index page
    # for this section.

    $sect->print(&end_html());
    $sect->close();
  }

  # Finish up the index page

  $index->print(&end_html());
  $index->close();

  return 1;
}

=head2 $obj->index_plucene(Plucene::Simple)

Index a collection with I<Plucene::Simple>.

Plucene::Simple doesn't always play nicely with this one and there
are instances where the string text passed to P:S causes it to fail.

I don't completely understand why, but the nuts and bolts of the
code here is sound.

Returns true or false.

=cut

sub index_plucene {
  my $self = shift;
  my $idx  = shift;

  #

  $self->pubdate() =~ /^([-\d]+)/;
  my $date = $1;

  #
  
  foreach my $section ($self->{'__doc'}->findnodes("/book/part")) {
      
      my @articles = $section->findnodes(XPATH_ARTICLES);
      next unless (@articles);
      
      # Munge the section title
      
      # my $sectname = $section->find(qq(partinfo/title));
      # &_sectname(\$sectname);
   
      # This needs to be be massaged to do something
      # more intelligent with referenced articles.

      foreach my $article (grep {$_->nodeName() ne "xref"} @articles) {

	  my $id   = &_articleid($article);
	  my $data = &hr_articleinfo($article);

	  $data->{ text } = join("\n",@{ &_cleanarticle($article) });

	  $idx->add( $id => $data);
      }
  }

  return 1;
}

=head2 $obj->index_contextgraph(Search::ContextGraph)

Index a collection with I<Search::ContextGraph>.

=cut

sub index_contextgraph {
  my $self  = shift;
  my $graph = shift;

  require File::Temp;

  #

  foreach my $section ($self->{'__doc'}->findnodes("/book/part")) {
      
      my @articles = $section->findnodes(XPATH_ARTICLES);
      next unless (@articles);
   
      # This needs to be be massaged to do something
      # more intelligent with referenced articles.

      foreach my $article (grep {$_->nodeName() ne "xref"} @articles) {

	  my ($fh,$fn) = File::Temp::tempfile(undef,UNLINK=>0);
	  $fh->print(join("\n",@{ &_cleanarticle($article) }));
	  $fh->close();

	  my $id  = &_uid($article);
	  my $uri = $self->path()."#xpointer(*//article[\@id='$id'])";

	  $graph->add_file($fn,name=>$uri);
	  unlink($fn);
      }
  }

  return 1;
}

=head2 $obj->index_delicious(ASCOPE::Delicious)

Index a collection with I<ASCOPE::Delicious>.

=cut

sub index_delicious {
    my $self      = shift;
    my $delicious = shift;

    #

    foreach my $section ($self->{'__doc'}->findnodes("/book/part")) {

	my @articles = $section->findnodes(XPATH_ARTICLES);
	next unless (@articles);

	foreach my $article (grep {$_->nodeName() ne "xref"} @articles) {
	  
	    my $id   = $article->getAttribute(qq(id));
	    my $info = &_articleinfo($article);

	    if (! $info) {
		next;
	    }

	    my $title    = $info->find(qq(title));
	    my $abstract = $info->find(qq(abstract/para));

	    my @tags = ();

	    #

	    foreach my $term (RELS_INDEXTERMS) {
		map {
		    push @tags, join("/",RELS_LABELS->{$term},&_normalize($_));
		} grep { 
		    $_ =~ /\w/;
		} &_indexterm($info,$term);
	    }
	    
	    foreach my $kw (RELS_KEYWORDS) {
		map { 
		    push @tags, join("/",RELS_LABELS->{$kw},&_normalize($_));
		} grep { 
		    $_ =~ /\w/;
		} &_keywords($info);
	    }

	    #

	    if (! @tags) {
		next;
	    }

	    #

	    my $bm = $delicious->add_post({url         => $id,
					   title       => $title,
					   description => $abstract,
					   tag         => join(" ",@tags)});
	}
	
    }

    return 1;
}

=head2 $obj->stats()

Returns a hash reference of relations and their number of referents
organized by topic. e.g. :

 geo => %
     Baltimore (Md) => 1
     Brazil => 1
     Brooklyn (NYC) => 1
     ...
 org => %
     Air Canada => 1
     Al Jazeera => 1
     Al Qaeda => 2

This method shares a degree of overlap with the I<weight_relations>
method that remains to be reconciled.

=cut

sub stats {
  my $self = shift;
  my %stats = ();

  #
  
  foreach my $section ($self->{'__doc'}->findnodes("/book/part")) {
      
      my @articles = $section->findnodes(XPATH_ARTICLES);
      next unless (@articles);

      foreach my $article (grep {$_->nodeName() ne "xref"} @articles) {
	  
	  my $info = &_articleinfo($article);

	  if (! $info) {
	      next;
	  }

	  #

	  foreach my $term (RELS_INDEXTERMS) {
	      map {  
		  $stats{ $term }->{ $_ } ++;
	      } grep { 
		  $_ =~ /\w/;
	      } &_indexterm($info,$term);
	  }

	  foreach my $kw (RELS_KEYWORDS) {
	      map {
		  $stats{ $kw }->{ $_ } ++;
	      } grep { 
		  $_ =~ /\w/;
	      } &_keywords($info);
	  }

      }

  }

  return \%stats;
}

=head2 $obj->relations()

Returns a hash ref of all the relations and their topics and
referents. e.g. :

 Johnson, Magic => %
       keywords => %
            Automobile Racing => 1
       org => %
            National Assn of Stock Car Auto Racing => 1
 Leiter, Al => %
       keywords => %
            Baseball => 1
       org => %
            New York Mets => 1

=cut

sub rels { return shift->relations(); }

sub relations {
  my $self = shift;
  my %rels = ();

  #
  
  foreach my $section ($self->{'__doc'}->findnodes("/book/part")) {
            
      my @articles = $section->findnodes(XPATH_ARTICLES);
      next unless (@articles);

      foreach my $article (grep {$_->nodeName() ne "xref"} @articles) {
	  
	  my %stats = ();
	  my $info  = &_articleinfo($article);

	  if (! $info) {
	      next;
	  }

	  #

	  foreach my $term (RELS_INDEXTERMS) {
	      map {  
		  $stats{ $term }->{ $_ } ++;
	      } grep { 
		  $_ =~ /\w/;
	      } &_indexterm($info,$term);
	  }

	  foreach my $kw (RELS_KEYWORDS) {
	      map { 
		  $stats{ $kw }->{ $_ } ++;
	      } grep { 
		  $_ =~ /\w/;
	      } &_keywords($info);
	  }

	  #

	  foreach my $topic (keys %stats) {
	      foreach my $what (keys %{$stats{$topic}}) {

		  foreach my $_topic (keys %stats) {
		      foreach my $_what (keys %{$stats{$_topic}}) {

			  next if (($_topic eq $topic) && ($_what eq $what));
			  $rels{ $topic }->{$what}->{$_topic}->{$_what} ++;
		      }
		  }
	      }
	  }

	  #

      }
  }

  #

  return \%rels;
}

=head2 $obj->weigh_relations(\%relations)

Returns a hash reference of all topics and the number of
references pointing to each.

  my $relations = $collection->relations();
  my $weights   = $collection->weigh_relations($relations);

The hash reference contains a nested hash whose key is I<__topic>
and which maps topic name to topic type. e.g. :

 %
     Accounting and Accountants => 3
     Air Canada => 2
     Air Pollution => 3
     Airlines and Airplanes => 2
     Al Jazeera => 4
     ...
 __topics => %
     Taxation => keywords 
     ...

This method shares a degree of overlap with the I<stats> method
that remains to be reconciled.

=cut

sub weigh_relations {
    my $self = shift;
    my $rels = shift;

    my %weights = ();

    foreach (keys %$rels) {
	my $topic = $rels->{$_};

	foreach (keys %$topic) {
	    my $what = $topic->{$_};

	    foreach my $rel (keys %$what) {
		my $type = $what->{$rel};

		foreach my $this (keys %$type) {
		    $weights{ $this} ++;
		    $weights{"__topic"}->{ $this } = $rel;		    
		}
	    }
	}
    }

    return \%weights;
}

=head2 $obj->heaviest_relations(\%weights)

Returns an array reference of one, or more, topics with
the highest number of references.

  my $relations = $collection->relations();
  my $weights   = $collection->weigh_relations($relations);
  my $heaviest  = $collection->heaviest_relations($weights);

=cut

sub heaviest_relations {
    my $self    = shift;
    my $weights = shift;

    my %foo = ();

    map {
	push @{ $foo{ $weights->{$_} } }, $_;
    } keys %$weights;
    
    my $i = (sort {$b <=> $a} keys %foo)[0];
    
    #

    return $foo{$i};
}

=head2 $obj->graph_relations(GraphViz,\@subset)

Graph all relations, or a user-defined subset of keywords or indexterms 
passed as an array reference.

    my $gv = GraphViz->new(...);

  my $relations = $collection->relations();
  my $weights   = $collection->weigh_relations($relations);
  my $heaviest  = $collection->heaviest_relations($weights);

  $collection->graph_relations($gv,$heaviest);
  $gv->as_svg(...);

=cut

sub graph_relations {
    my $self  = shift;
    my $graph = shift;
    my $slice = shift;
    my $args  = shift;

    $args ||= {};

    if ($slice) {
	return $self->_graph_slice($graph,$slice,$args);
    }

    #

    my $relations = $self->relations();
    my $weights   = $self->weigh_relations($relations);

    #

    foreach my $topic (keys %$relations) {
	foreach my $what (keys %{$relations->{$topic}}) {

	    my %node_args = (fontsize  => ($args->{fontsize}  || 8),
			     shape     => ($args->{shape}     || RELS_SHAPES->{$topic}),
			     fillcolor => ($args->{fillcolor} || RELS_COLOURS->{$topic}),
			     style     => "filled",);

	    if ($args->{nolabels}) {
		$node_args{ label } = "";
	    }
	    
	    $graph->add_node($what,%node_args);
	}
    }

    #

    foreach (keys %$relations) {
	my $topic = $relations->{$_};

	foreach my $node (keys %$topic) {
	    my $what = $topic->{$node};

	    foreach (keys %$what) {
		my $type = $what->{$_};

		foreach my $edge (keys %$type) {

		    $graph->add_edge($node    => $edge,
				     weight   => $weights->{ $edge },
				     samehead => 1,
				     sametail => 1,
				     minlen   => 2);
		}
	    }
	}
    }

    #

    return 1;
}

sub _graph_slice {
    my $self  = shift;
    my $graph = shift;
    my $which = shift;
    my $args  = shift;

    my $rels    = $self->relations();
    my $weights = $self->weigh_relations($rels);

    #

    my %nodes   = ();

    #

    foreach (keys %$rels) {
	my $topic = $rels->{$_};

	foreach my $this (keys %$topic) {

	    # This sometimes fails (dunno) :
	    # next unless (grep /^($this)$/,@$which);

	    # So we are reduced to this :

	    my $ok = 0;

	    map { 
		if ($this eq $_) {
		    $ok = 1;
		    last;
		}
	    } @$which;

	    if (! $ok) {
		next;
	    }

	    #

	    foreach my $rel (keys %{$topic->{$this}}) {
		$nodes{ $this }->{ $rel } = $topic->{$this}->{$rel};
	    }
	}
    }

    #

    foreach my $first (keys %nodes) {

	my %node_args = (fontsize  => 6,
			 shape     => "ellipse",
			 fillcolor => RELS_COLOURS->{ $weights->{'__topic'}->{$first} },
			 style     => "filled");
	
	if ($args->{nolabels}) {
	    $node_args{ label } = "";
	}

	$graph->add_node($first,%node_args);

	foreach my $rel (keys %{$nodes{$first}}) {

	    foreach my $what (keys %{$nodes{$first}->{$rel}}) {

		my %node_args = (fontsize  => 6,
				 shape     => RELS_SHAPES->{$rel},
				 fillcolor => RELS_COLOURS->{$rel},
				 style     => "filled");
		
		if ($args->{nolabels}) {
		    $node_args{ label } = "";
		}

		$graph->add_node($what,%node_args);
	    }
	}
    }

    # 

    foreach my $first (keys %nodes) {
	foreach my $rel (keys %{$nodes{$first}}) {
	    foreach my $edge (keys %{$nodes{$first}->{$rel}}) {

		$graph->add_edge($first   => $edge,
				 samehead => 1,
				 sametail => 1,
				 minlen   => 2);

		foreach my $topic (keys %{$rels->{$rel}->{$edge}}) {

		    foreach my $what (keys %{$rels->{$rel}->{$edge}->{$topic}}) {

			if (exists($nodes{$first}->{$topic}->{$what})) {
			    $graph->add_edge($edge    => $what,
					     samehead => 1,
					     sametail => 1,
					     minlen   => 2);
			}


		    }
		}
	    }
	}
    }

    #

    return 1;
}

=head2 $obj->hr_articleinfo(XML::LibXML::Node)

Utility method to return the I<articleinfo> for an article
as a hash reference.

=cut

sub hr_articleinfo {
    my $article = shift;

    my %data = ();

    # Pull out the articleinfo and create some
    # variables that we'll use below.
    
    my $info = &_articleinfo($article);

    if (! $info) {
	return \%data;
    }

    #

    if (my $date = $info->find(qq(pubdate))) {
	$date =~ /^(\d{4}-\d{2}-\d{2}).*/;
	$data{ date } = $1;
    }

    if (my $title = $info->find(qq(title))) {
	$title =~ s/\n/ /gm;
	$data{ title } = $title;
    }
    
    if (my $author = &_author($info)) {
	$data{ author } = "$author";
    }
    
    if (my $person = join(",",&_indexterm($info,"per"))) {
	$data{ person } = $person;
    }
    
    if (my $org = join(",",&_indexterm($info,"org"))) {
	$data{ org } = $org;
    }
    
    if (my $geo = join(",",&_indexterm($info,"geo"))) {
	$data{ geo } = $geo;
    }
    
    if (my $kw = join(",",&_keywords($info))) {
	$data{ keywords } = $kw;
    }

    return \%data;
}

sub _normalize {
    my $str = shift;

    $str = lc($str);

    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    $str =~ s/[^a-z\d\s]//g;

    $str =~ s/\s\s+/ /g;
    $str =~ s/\s/_/g;

    $str =~ s/^(a|an|the)_(.*)$/$2_$1/;
    return $str;
}

sub _articleinfo {
    my $article = shift;

    my $info = $article->getChildrenByTagName(qq(articleinfo))->[0];
    
    if (! $info) {
	my $id = $article->getAttribute(qq(id));

	warn "can't get article info for article '$id'\n";
	return undef;
    }

    return $info;
}

sub _indexterm {
    my $info = shift;
    my $term = shift;

    return map { 
	$_->string_value();
    } $info->find(qq(indexterm[\@type='$term']/primary));
}

sub _keywords {
    my $info = shift;

    return map { 
	$_->string_value();
    } $info->find(qq(keywordset/keyword))->get_nodelist();
}

sub _author {
    my $info = shift;

    if (my $author = $info->find(qq(authorgroup/corpauthor))) {
	return $author;
    }

    return join(",",map { 
	join(" ",$_->find(qq(firstname)),$_->find(qq(surname)));
    } $info->find(qq(authorgroup/author))->get_nodelist());
}

sub _uid {
    my $article = shift;
    return $article->getAttribute(qq(id));
}

sub _articleid {
    my $article = shift;

    if (my $id = $article->find(qq(articleinfo/indexterm[\@type='articleid']))) {
	return $id->[0]->find(qq(primary));
    }

    return md5_hex(&_uid($article));
}

sub _cleanarticle {
    my $article = shift;

    return [ map {
	$_ =~ s/^\s+//; $_;
    } grep { 
	$_ =~ /\w/;
    } map {
	$_->textContent();
    } ($article->getChildrenByTagName(qq(para))) ];
}

sub _sectname {
  my $sr_name = shift;
  $$sr_name = lc($$sr_name);
  $$sr_name =~ s/ //g;
  $$sr_name =~ s/:/\//g;
}

=head1 VERSION

1.1

=head1 DATE

$Date: 2004/07/10 17:04:12 $

=head1 AUTHOR

Aaron Straup Cope E<lt>ascope@cpan.orgE<gt>

=head1 SEE ALSO

L<Plucen::Simple>

L<Search::ContextGraph>

L<GraphViz>

L<ASCOPE::Delicious>

L<XML::LibXML>

=head1 LICENSE

Copyright (c) 2003-2004, Aaron Straup Cope. All Rights Reserved.

This is free software, you may use it and distribute it under
the same terms as Perl itself.

=cut

return 1;
